/* SPDX-License-Identifier: GPL-2.0
 *
 * Copyright (C) 2020 Jason A. Donenfeld. All Rights Reserved.
 */

#include <windows.h>

#pragma code_page(65001) // UTF-8

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST manifest.xml

#define STRINGIZE(x) #x
#define EXPAND(x) STRINGIZE(x)

VS_VERSION_INFO VERSIONINFO
FILEOS         VOS_NT_WINDOWS32
FILETYPE       VFT_APP
FILESUBTYPE    VFT2_UNKNOWN
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904b0"
    BEGIN
      VALUE "CompanyName", "WireGuard LLC"
      VALUE "FileDescription", "WireGuard wg(8) CLI: Fast, Modern, Secure VPN Tunnel"
      VALUE "FileVersion", EXPAND(VERSION_STR)
      VALUE "InternalName", "wg"
      VALUE "LegalCopyright", "Copyright © 2015-2020 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved."
      VALUE "OriginalFilename", "wg.exe"
      VALUE "ProductName", "WireGuard"
      VALUE "ProductVersion", EXPAND(VERSION_STR)
      VALUE "Comments", "https://www.wireguard.com/"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x409, 0x4b0
  END
END
